/* pumpexs.c - client and server program that uses the EXS socket interface
 *		 This program accepts a number of command line options, then
 *		 sends that number of bytes from the client to the server.
 *  on Solaris this program must be linked with -lnsl and -lsocket
 */

/*
 * Copyright 2007, 2008, 2009, 2010, 2011, 2012, 2013, 2014 by
 *	University of New Hampshire Interoperability Laboratory
 *	121 Technology Drive, Suite 2
 *	Durham, NH 03824-4716
 *	phone: 1-603-862-0090
 *	fax: 1-603-862-4181
 *	e-mail: info@iol.unh.edu
 * All rights reserved.
 *
 * This software is available to you under a choice of one of two
 * licenses.  You may choose to be licensed under the terms of the GNU
 * General Public License (GPL) Version 2, available from the file
 * COPYING.GPL in the main directory of this source tree, or the
 * BSD license, available from the file COPYING.BSD in the main directory
 * of this source tree.
 *
 * This material is based upon work supported by the National Science
 * Foundation under Grant No. OCI-1127228.
 *
 * Any opinions, findings, and conclusions or recommendations
 * expressed in this material are those of the authors and do not
 * necessarily reflect the views of the National Science Foundation.
 */

/*
 * The UNH IOL implementation of the Extended Sockets API (ES-API) Issue 1.0
 *
 * defined in:	The Interconnect Software Consortium Issue 1.0
 *		ISBN: 1-931624-52-6
 *		Published by The open Group, January 2005.
 *
 * all page references in this source code are to pages in that document.
 */

/*
 * Most of the coding in this file done by:
 *	Bob Russell
 *	rdr@iol.unh.edu
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#define _POSIX_C_SOURCE 200112L
#define _ISOC99_SOURCE
#define __EXTENSIONS__
#define _XOPEN_SOURCE 600

#include <assert.h>
#include <limits.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>
#include <arpa/inet.h>
#include <sys/socket.h>
#include <ctype.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <fcntl.h>
#include <time.h>
#include <pthread.h>

#include "utilities.h"

#define TEXT_SIZE	1024			/* size of text buffer */

#define MEGA_BYTES	(1000 * 1000)
#define MEGA_UNITS	(1000000.0)		/* microseconds per second */

#define ADDRESS_ITEMS	1
#define LISTENING_DEPTH	3
#define SEND_DEPTH	2
#define RECV_DEPTH	2

#define OPTIONS		"a:A:p:P:c:C:s:S:g:G:hHdDioOb:B:uUm:M:nNkKxXqQl:L:r:R:tTeEf:F:I:w:W:yYz:Z:"
#define WHITE_SPACE			" \t\n\r\f\v"
#define DEFAULT_PORT			"54321"
#define DEFAULT_COUNT			16
#define DEFAULT_SIZE			256
#define DEFAULT_NBUFS			1
#define DEFAULT_N_LISTENER_LOOPS	UINT_MAX

enum suboptions {
	SUBOPT_NO_STREAM_ADVERT,
	SUBOPT__MAX,
};

char *const suboptions[] = {
	[SUBOPT_NO_STREAM_ADVERT] = "no_stream_advert",
	NULL,
};

struct user_ahandle	{
	size_t		length;
	void		*address;
	exs_mhandle_t	mhandle;
	};

struct child_params	{
	int		client_fd;
	struct sockaddr_storage	client_addr;
	};


/* Mutex to guard the exit condition and number of child threads created */
static pthread_mutex_t exit_mutex = PTHREAD_MUTEX_INITIALIZER;

/* Guarded by: exit_mutex */
static int nthreads = 0;

/* Condition: nthreads == 0; guarded by: exit_mutex */
static pthread_cond_t exit_cond = PTHREAD_COND_INITIALIZER;

static char *my_name = "pumpexs";


/*	-a: server_name == IP address of server	*/
static char			*server_name;

/*	-b: nbufs == number of simultaneously active buffers	*/
static unsigned int		nbufs;

/*	-c: count == max number of transactions	*/
static unsigned int		max_count;

/*	-g: sg_elements == number of scatter_gather elements in exs_sendfile()*/
static int			sg_elements = 0;

/*	-i: input_flag == 1 for server (reader), 0 for client (writer)	*/
static int			input_flag = 0;

/*	-l: n_listener_loops == number of listener loops	*/
static unsigned int		n_listener_loops;

/*	-o: output_flag == 1 for client (writer), 0 for server (reader)	*/
static int			output_flag = 0;

/*	-p: server_port == port number of server */
static char			*server_port;

/*	-s: size == number of bytes per transaction */
static int			size;

/*	-t: socktype == SOCK_STREAM (TCP semantics), default is SOCK_SEQPACKET*/
static int			socktype = SOCK_SEQPACKET;

/*	-u: unregistered_buffers_flag = 1 to use unregistered buffers for I/O */
static int			unregistered_buffers_flag = 0;

/*	Currently not used
 */
static int			recv_flags = 0;

/*	Currently not used
 */
static int			send_flags = 0;


/*	-q: do not wait for notification after send or recv (queue is null) */
static int			no_queue = 0;

/*	-y: verify */
static int			verify = 0;

static size_t			align = 64;

static struct conn_params	our_conn_params;


struct control_block {
	/*	pointers to malloced buffers, one for each transaction */
	void			**buftab;

	/*	handles for registering buffers	*/
	exs_mhandle_t		*buf_mhantab;

	/*	handles for user ahandle structures */
	exs_ahandle_t		*buf_ahantab;

	/*	scatter_gatter lists for buffers */
	exs_xferfile_t		*xfertab;

	/*	queue for sending messages to remote side	*/
	exs_qhandle_t		send_queue;

	/*	queue for receiving messages from remote side	*/
	exs_qhandle_t		recv_queue;

	/*	total number of bytes actually transfered	*/
	long long unsigned	total_bytes;

	/*	total number of transfers actually done	*/
	long long unsigned	total_transfers;

	/*	-c: count == number of transactions	*/
	unsigned int		count;

	/*	set to time transfers started	*/
	struct timespec		start_time;

	/*	set to usage when transfers started	*/
	struct rusage		start_usage;

	/*	work space for current time	*/
	int			secs, usecs;

	/*	work space for current usage	*/
	int			user_secs, user_usecs;
	int			kernel_secs, kernel_usecs;
};	/* struct control_block */


/* returns EXIT_SUCCESS or EXIT_FAILURE */
int
dealloc_control_block(struct control_block *cb, int level)
{
	unsigned int	n;
	/* sizeof(struct sockaddr_in)	is 16
	 * sizeof(struct sockaddr_in6)	is 28
	 * sizeof(struct sockaddr)	is 16 !!! should be 28 !!!
	 */
	void			**bptr;
	exs_mhandle_t		*mptr;
	exs_ahandle_t		*aptr;
	char			message[100];
//	struct timespec		small_delay;

	switch (level) {
default:
case 7:
	TRACE_DBG(LOG_APPLICATION, "%s exs_qdelete recv_queue %p",
			message, cb->recv_queue);
	if (exs_qdelete(cb->recv_queue) < 0) {
		fprintf(stderr, "%s %s: %s\n", message,
				"exs_qdelete recv_queue", strerror(errno));
		assert(level > 0);
		level--;
	}
case 6:
		TRACE_DBG(LOG_APPLICATION, "%s exs_qdelete send_queue %p",
			message, cb->send_queue);
	if (exs_qdelete(cb->send_queue) < 0) {
		fprintf(stderr, "%s %s: %s\n", message,
				"exs_qdelete send_queue", strerror(errno));
		assert(level > 0);
		level--;
	}
case 5:
	/* deregister and free the memory buffers and their memory handles */
	bptr = cb->buftab;	/* point at array of "nbufs" pointers to buffs*/
	mptr = cb->buf_mhantab;	/* point at array of "nbufs" memory handles */
	aptr = cb->buf_ahantab;	/* point at array of "nbufs" user ahandles */
	for (n = 0; n < nbufs; n++) {
		TRACE_DBG(LOG_APPLICATION,
			"%s deregistered %d bytes at buftab[%d] %p, buf_mhantab[%d] %p",
			message, size, n, *bptr, n, *mptr);
		if (unregistered_buffers_flag == 0) {
			if (*mptr == EXS_MHANDLE_INVALID || *mptr == NULL) {
				break;
			}
			if (exs_mderegister(*mptr, 0)) {
			fprintf(stderr, "%s %s: %s\n", message,
				"deregister buf_mhantab[n]", strerror(errno));
			if (level > 0) {
				level--;
			}
			}
		}
		if (*bptr == NULL) {
			break;
		}
		free(*bptr);
		bptr++;		/* point at next pointer to buffer */
		mptr++;		/* point at next memory handle */
		aptr++;		/* point at next user ahandle */
	}
	free(cb->xfertab);

	/* free the arrays of pointers to buffers and their memory handles */
case 4:
	free(cb->buf_ahantab);
case 3:
	free(cb->buf_mhantab);
case 2:
	free(cb->buftab);
case 1:
	free(cb);
case 0:
	break;
	}	/* switch */

	return level > 7 ? EXIT_SUCCESS : EXIT_FAILURE;
}	/* dealloc_control_block */


/* allocates and initializes a new struct control_block */
int
alloc_control_block(struct control_block **new_cb, char *message)
{
	int	j, k, chunk;
	int	alloc_chunk, alloc_size;
	/* sizeof(struct sockaddr_in)	is 16
	 * sizeof(struct sockaddr_in6)	is 28
	 * sizeof(struct sockaddr)	is 16 !!! should be 28 !!!
	 */
	unsigned int		n;
	char			*ptr;
	exs_xferfile_t		*xptr;
	void			**bptr;
	exs_mhandle_t		*mptr;
	exs_ahandle_t		*aptr;
//	struct timespec		small_delay;

	struct control_block	*cb;

	cb = calloc(1, sizeof(struct control_block));
	if (cb == NULL) {
		fprintf(stderr, "%s %s: %s\n", message,
				"control_block", strerror(errno));
		return dealloc_control_block(cb, 0);
	}

	/* need to allocate and register "nbufs" buffers,
	 * each of "size" bytes of memory, each 64-byte-aligned,
	 * that we will send to or receive from the server
	 * buftab holds the "nbufs" pointers to these buffers
	 * buf_mhantab holds the "nbufs" memory handles for these buffers
	 * buf_ahantab holds "bufs" ptrs to ahandle structures for these buffers
	 */
	if ((cb->buftab = calloc(nbufs, sizeof(void *))) == NULL) {
		fprintf(stderr, "%s %s: %s\n", message,
					"buftab", strerror(errno));
		return dealloc_control_block(cb, 1);
	}
	if ((cb->buf_mhantab = calloc(nbufs, sizeof(exs_mhandle_t))) == NULL) {
		fprintf(stderr, "%s %s: %s\n", message,
					"buf_mhantab", strerror(errno));
		return dealloc_control_block(cb, 2);
	}

	if ((cb->buf_ahantab = calloc(nbufs, sizeof(exs_ahandle_t))) == NULL) {
		fprintf(stderr, "%s %s: %s\n", message,
					"buf_ahantab", strerror(errno));
		return dealloc_control_block(cb, 3);
	}

	/* k is number of chunks in one buffer,
	 * chunk is number of bytes in 1 chunk */
	if ((k = sg_elements) <= 0)
		k = 1;
	chunk = size / k;
	if (chunk <= 0) {
		/* user is asking for too many chunks in one buffer */
		k = size;
		chunk = 1;
	}
	/* allocate chunks in multiples of 64 bytes */
	alloc_chunk = (chunk + 63) & (~63);
	alloc_size = alloc_chunk * k;

	/* nbufs rows in xfertab, one for each buffer,
	 * k columns in xfertab, one for each chunk in a buffer
	 * (nbufs * k) elements in xfertab, each of type exs_xferfile_t
	 */
	if ((cb->xfertab = calloc(nbufs * k, sizeof(exs_xferfile_t))) == NULL) {
		fprintf(stderr, "%s %s: %s\n", message,
					"xfertab", strerror(errno));
		return dealloc_control_block(cb, 4);
	}
	TRACE_DBG(LOG_APPLICATION,
		"%s allocated %lu bytes for xfertab %p, %u buffers, each with %d chunks of %lu bytes",
		message, (unsigned long)(nbufs * k * sizeof(exs_xferfile_t)),
		cb->xfertab, nbufs, k, (unsigned long)sizeof(exs_xferfile_t));

	xptr = cb->xfertab;	/* point at array of (nbufs*k) exs_xferfile_t */
	bptr = cb->buftab;	/* point at array of "nbufs" pointers to buffs*/
	mptr = cb->buf_mhantab;	/* point at array of "nbufs" memory handles */
	aptr = cb->buf_ahantab;	/* point at array of "nbufs" user ahandles */
	for (n = 0; n < nbufs; n++) {

		/* allocate the next buffer on a page boundary */
		if (posix_memalign(bptr, sysconf(_SC_PAGE_SIZE), alloc_size)) {
			fprintf(stderr, "%s No space for %d-byte buffer[%d]\n",
				message, alloc_size, n);
			return dealloc_control_block(cb, 5);
		}

		if (unregistered_buffers_flag) {
			*mptr = EXS_MHANDLE_UNREGISTERED;
			fprintf(stderr, "%s %d bytes of memory at "
				"buftab[%d] %p, buf_mhantab[%d] %p "
				"(EXS_MHANDLE_UNREGISTERED)\n",
				message, size, n, *bptr, n, *mptr);
		} else {
			*mptr = exs_mregister(*bptr,alloc_size,EXS_ACCESS_ALL);
			if (*mptr == EXS_MHANDLE_INVALID) {
				fprintf(stderr, "%s %s: %s\n", message,
					"exs_mregister message buffer",
					strerror(errno));
				return dealloc_control_block(cb, 5);
			}
			TRACE_DBG(LOG_APPLICATION,
				"%s registered %d bytes at buftab[%d] %p, buf_mhantab[%d] %p",
				message, alloc_size, n, *bptr, n,*mptr);
		}

		/* allocate the next user ahandle and fill it in */
		if ((*aptr = malloc(sizeof(struct user_ahandle))) == NULL) {
			fprintf(stderr, "%s No space for %d-byte user "
				"ahandle[%d]\n", message,
				(int)sizeof(struct user_ahandle), n);
			return dealloc_control_block(cb, 5);
		}
		(*(struct user_ahandle **)aptr)->length = size;
		(*(struct user_ahandle **)aptr)->address = *bptr;
		(*(struct user_ahandle **)aptr)->mhandle = *mptr;

		TRACE_DBG(LOG_APPLICATION,
			"%s allocated %d bytes for ahandle[%d] %p, length %d, address %p, mhandle %p",
			message, (int)sizeof(struct user_ahandle), n,
			*aptr,
			(int)(*(struct user_ahandle **)aptr)->length,
			(*(struct user_ahandle **)aptr)->address,
			(*(struct user_ahandle **)aptr)->mhandle);

		/* fill in the sg-element scatter-gather list for this buffer
		 * each buffer is chopped into k equal sized chunks
		 */
		ptr = *bptr;	/* point at first chunk in this buffer */
		for (j = 0; j < k; j++) {
			TRACE_DBG(LOG_APPLICATION,
				"%s &xfertab[%d][%d] %p: &iov_base %p, &iov_len %p, &iov_mhandle %p",
				message, n, j, xptr,
				&xptr->exs_xfv.exs_iov.iov_base,
				&xptr->exs_xfv.exs_iov.iov_len,
				&xptr->exs_xfv.exs_iov.iov_mhandle);
			xptr->exs_xfv_type = EXS_IOVEC;
			xptr->exs_xfv.exs_iov.iov_base = ptr;
			xptr->exs_xfv.exs_iov.iov_len = chunk;
			xptr->exs_xfv.exs_iov.iov_mhandle = *mptr;
			TRACE_DBG(LOG_APPLICATION,
				"%s xfertab[%d][%d]: iov_base %p, iov_len %zu, iov_mhandle %p",
				message, n, j,
				xptr->exs_xfv.exs_iov.iov_base,
				xptr->exs_xfv.exs_iov.iov_len,
				xptr->exs_xfv.exs_iov.iov_mhandle);
			/* point at next allocated chunk in this buffer */
			ptr += alloc_chunk;
			xptr++;		/* point at next exs_xferfile_t */
		}
	bptr++;		/* point at next pointer to buffer */
	mptr++;		/* point at next memory handle */
	aptr++;		/* point at next user ahandle */
	}

	/* create exs event queues for sending messages to and
	 * receiving replies back from the remote side
	 */
	cb->send_queue = exs_qcreate(nbufs + 1);
	if (cb->send_queue == EXS_QHANDLE_INVALID) {
		fprintf(stderr, "%s %s: %s\n", message,
				"exs_qcreate send queue", strerror(errno));
		return dealloc_control_block(cb, 5);
	}
	TRACE_DBG(LOG_APPLICATION, "%s created send_queue %p of depth %u",
			message, cb->send_queue, nbufs + 1);

	cb->recv_queue = exs_qcreate(nbufs + 1);
	if (cb->recv_queue == EXS_QHANDLE_INVALID) {
		fprintf(stderr, "%s %s: %s\n", message,
				"exs_qcreate recv queue", strerror(errno));
		return dealloc_control_block(cb, 6);
	}
	TRACE_DBG(LOG_APPLICATION,
			"%s created recv_queue %p of depth %u",
			message, cb->recv_queue, nbufs + 1);

	*new_cb = cb;
	return EXIT_SUCCESS;
}	/* alloc_control_block */


void
print_rates( struct control_block *cb, unsigned int buffers, char *message)
{
	double			dsecs, dusecs, dbytes, dtransfers;

	/* stop the timers immediately after last transaction completes */
	get_relative_time(&cb->start_time, &cb->secs, &cb->usecs, message);
	get_relative_usage(&cb->start_usage, &cb->user_secs, &cb->user_usecs,
				&cb->kernel_secs, &cb->kernel_usecs, message);
	fprintf(stderr, "%s waited for %u buffers "
		"at %d.%06d\n", message, buffers, cb->secs, cb->usecs);

	fprintf(stderr, "%s %18llu transfers\n", message, cb->total_transfers);
	fprintf(stderr, "%s %18llu bytes %s\n", message, cb->total_bytes,
		"transfered");
	fprintf(stderr, "%s %11d.%06d seconds elapsed\n", message,
			cb->secs, cb->usecs);
	fprintf(stderr, "%s %11d.%06d seconds user (%.4f%%)\n",
			message, cb->user_secs, cb->user_usecs,
			percentage(cb->secs, cb->usecs,
						cb->user_secs, cb->user_usecs));
	fprintf(stderr, "%s %11d.%06d seconds kernel (%.4f%%)\n",
			message, cb->kernel_secs, cb->kernel_usecs,
			percentage(cb->secs, cb->usecs, cb->kernel_secs,
						cb->kernel_usecs));
	dusecs = cb->secs;
	dusecs *= MEGA_UNITS;
	dusecs += cb->usecs;
	dsecs = dusecs / MEGA_UNITS;
	dbytes = ((double)cb->total_bytes);
	dtransfers = ((double)cb->total_transfers);
	if (dusecs != 0) {
		fprintf(stderr, "%s %18.3f transfers/second\n",
			message, dtransfers/dsecs);
		if (dtransfers != 0) {
			fprintf(stderr, "%s %18.3f microseconds/transfer\n",
				message, dusecs/dtransfers);
		}
		fprintf(stderr, "%s %18.3f Megabytes/second\t"
			"(%.0f bytes per Megabyte)\n",
			message, dbytes/dusecs, MEGA_UNITS);
		fprintf(stderr, "%s %18.3f Megabits/second\t"
			"(%.0f bits per Megabit)\n\n",
			message, (8.0*dbytes)/dusecs, MEGA_UNITS);
	}
}	/* print_rates */


static int
close_a_connection(struct control_block *cb, int fd, char* message)
{
	int		len;
	exs_event_t	completion;
	struct timeval	timeout;

	timeout.tv_sec = 3;
	timeout.tv_usec = 500000;	/* half a second in microseconds */
	TRACE_DBG(LOG_APPLICATION,
		"%s close_a_connection calling exs_close, fd %d", message, fd);
	if ((len = exs_close(fd, 0, cb->recv_queue, NULL)) < 0) {
		fprintf(stderr, "%s exs_close: %s\n", message, strerror(errno));
	} else {
		do {
			len = asynch_wait(cb->recv_queue, &completion, &timeout);
			if (len < 0) {
				fprintf(stderr, "%s %s: %s: %s\n", message,
						"close_a_connection asynch_wait",
						exs_evt_type_to_string(
							completion.exs_evt_type),
						strerror(errno));
				len = -1;
			} else if (completion.exs_evt_type != EXS_EVT_CLOSE) {
				fprintf(stderr, "%s %s: Unexpected event: (%d) %s\n",
						message,
						"close_a_connection asynch_wait",
						completion.exs_evt_type,
						exs_evt_type_to_string(
							completion.exs_evt_type));
			}
		} while (completion.exs_evt_type != EXS_EVT_CLOSE);
	}
	TRACE_DBG(LOG_APPLICATION,
		"%s close_a_connection returning len %d, fd %d",
		message, len, fd);
	return len;
}	/* close_a_connection */


static void
print_sock_sizes(char *message)
{
	fprintf(stderr, "%s %32s %lu\n", message,
		"sizeof(struct sockaddr)", sizeof(struct sockaddr));
	fprintf(stderr, "%s %32s %lu\n", message,
		"sizeof(struct sockaddr_in)", sizeof(struct sockaddr_in));
	fprintf(stderr, "%s %32s %lu\n", message,
		"sizeof(struct sockaddr_in6)", sizeof(struct sockaddr_in6));
	fprintf(stderr, "%s %32s %lu\n", message,
		"sizeof(struct sockaddr_storage)",
		sizeof(struct sockaddr_storage));
}	/* print_sock_sizes */


/* returns 0 on success, -1 on error */
int
client_task( void )
{
	struct sockaddr_storage	client_addr;
	struct sockaddr_storage	server_addr;
	int			fd, len, len2, j;
	unsigned int		i, kk, m, n, failed;
	char			*ptr, nextch;
	const char		*endptr;
	exs_event_t		completion;
	void			*bufptr;
	exs_mhandle_t		buf_handle;
#if 0
	struct timespec		small_delay;
#endif
	exs_xferfile_t		*xptr;
	char			*text_buf;
	char			message[100];
	exs_qhandle_t		local_send_queue;
	struct control_block	*cb;
	int			level = 7;

	if ((len = posix_memalign((void **)&text_buf, align, TEXT_SIZE + 2)) != 0) {
		fprintf(stderr, "Could not allocate initial buffer: %s\n",
				strerror(len));
	}

	failed = 0;
	sprintf(message, "%s [%lu] %s:",
		my_name, pthread_self(), __func__);

	print_sock_sizes(message);

	fprintf(stderr, "%s Enter\n", message);

	if (alloc_control_block(&cb, message) == EXIT_FAILURE) {
		free(text_buf);
		return -1;
	}

	/* establish a client connection to indicated server */
	TRACE_DBG(LOG_APPLICATION,
		"%s calling openclient(port=%s, name=%s) to connect to server",
		message, server_port, server_name);
	fd = openclient(server_port, server_name, socktype,
			(struct sockaddr *)&server_addr,
			(struct sockaddr *)&client_addr,
			&our_conn_params);
	TRACE_DBG(LOG_APPLICATION,
			"%s client's openclient returned fd %d",
			message, fd);
	if (fd < 0)
		goto err0;

	sprintf(message, "%s [%lu] %s fd %d:",
		my_name, pthread_self(), "client", fd);

	/* we are now successfully connected to a remote server */
	print_ip_address((struct sockaddr *)&client_addr, "client",
			message);
	print_ip_address((struct sockaddr *)&server_addr, "server",
			message);
	print_attributes(fd, message);
	exs_print_connection_attributes(stderr, fd, message);

	if (no_queue)
		local_send_queue = NULL;
	else
		local_send_queue = cb->send_queue;

#if 0
	fprintf(stderr,"%s client giving server 0.5 seconds to set up\n",
								message);
	small_delay.tv_sec = 0;
	small_delay.tv_nsec = 500000000;
	if (nanosleep(&small_delay, NULL))
		fprintf(stderr, "%s %s: %s\n", message,
				"nanosleep small delay", strerror(errno));
#endif

	/* send size bytes of printable data to server
	 * in each of nbufs buffers
	 * do this "count" times
	 */
	cb->count = max_count;
	nextch = '0';
	if (!verify) {
		/* prepopulate the data since we are not verifying it */
		for (n = 0; n < nbufs; n++) {
			nextch = '0' + (n % 10);
			ptr = cb->buftab[n];
			for (j = 0; j < size; j++) {
				if (nextch == 0x7f)
					nextch = ' ';
				*ptr++ = nextch++;
			}
		}
	}

	fprintf(stderr, "%s writing %d-byte buffers %u times "
		"with %u active simultaneously\n",
		message, size, max_count, nbufs);

	/* m gets min(max_count, nbufs) */
	m = max_count;
	if (m > nbufs)
		m = nbufs;

	/* send a "go" to server to say we're ready */
	strcpy(text_buf, "go");
	len = strlen(text_buf);
	fprintf(stderr, "%s writeblock len %d \"%s\"\n",
		message, len, text_buf);
	len = writeblock(fd, text_buf, align, 0,
				EXS_MHANDLE_UNREGISTERED, cb->send_queue);
	if (len <= 0) {
		fprintf(stderr, "%s writeblock %s\n",
			message, strerror(errno));
		goto err1;
	}

	/* wait for server to tell us "ok" */
	len = readpacket(fd, text_buf, TEXT_SIZE,
				EXS_MHANDLE_UNREGISTERED, cb->recv_queue);
	if (len < 0) {
		fprintf(stderr, "%s readpacket %s\n",
			message, strerror(errno));
		goto err1;
	} else if (len == 0) {
		fprintf(stderr, "%s readpacket got EOF\n", message);
	}
	text_buf[2] = '\0';
	fprintf(stderr, "%s readpacket len %d \"%s\"\n",
		message, len, text_buf);

	len = 0;

	/* start the timers immediately before first transaction starts */
	if( getrusage(RUSAGE_SELF, &cb->start_usage) < 0 )
		{
		fprintf(stderr, "%s getrusage start %s\n",
			message, strerror(errno));
		goto err1;
		}
	if( clock_gettime(CLOCK_EXS, &cb->start_time) < 0 )
		{
		fprintf(stderr, "%s gettimeofday start %s\n",
			message, strerror(errno));
		goto err1;
		}

	/* send out the first nbufs buffers without waiting */
	for (i = 0, n = 0; i < m; i++, n++) {
		TRACE_DBG(LOG_APPLICATION,
			"%s starting iteration %u, fd %d, len %d, buffer address %p, buf_handle %p, local_send_queue %p, total_bytes %llu",
			message, i, fd, size,
			cb->buftab[n], cb->buf_mhantab[n],
			local_send_queue, cb->total_bytes);
		if (verify) {
			ptr = (char *)cb->buftab[n];
			endptr = &ptr[size];
			while (ptr != endptr) {
				if (nextch == 0x7f)
					nextch = ' ';
				*ptr++ = nextch++;
			}
			*ptr = '\0';
		}
		if (sg_elements) {
			len = asynch_sendfile(fd, &cb->xfertab[n], sg_elements,
				local_send_queue, &cb->xfertab[n]);
		} else {
			len = asynch_send(fd, (char *)cb->buftab[n], size,
					send_flags, cb->buf_mhantab[n],
					local_send_queue);
		}
		if (len <= 0) {
			fprintf(stderr, "%s %s: %s\n", message,
				"asynch_send", strerror(errno));
			cb->count = 0;	/* stop sending packets */
			failed = 1;
			break;
		}
		cb->total_transfers++;
	}	/* for */

	get_relative_time(&cb->start_time, &cb->secs, &cb->usecs, message);
	fprintf(stderr, "%s sent first %u buffers without waiting "
		"at %d.%06d\n", message, i, cb->secs, cb->usecs);

	/* send out remaining packets after waiting for previous completion
	 * in order to reuse its buffer and buffer registration
	 */
	kk = 0;
	xptr = NULL;
	buf_handle = NULL;
	bufptr = NULL;
	for ( ; i < cb->count; i++) {
		if (no_queue == 0)
		{
		len = asynch_wait(cb->send_queue, &completion, NULL);
		//if (i % 1000 == 0) fprintf(stderr, ".");

		TRACE_DBG(LOG_APPLICATION,
			"%s got completion type %d, expected %d, errno %d, fd %d, length %zu",
			message, completion.exs_evt_type,
			sg_elements ? EXS_EVT_SENDFILE
					    : EXS_EVT_SEND,
			completion.exs_evt_errno,
			completion.exs_evt_socket,
			(sg_elements
				?  completion.exs_evt_union
				       .exs_evt_xferfile.exs_evt_length
				: completion.exs_evt_union
				       .exs_evt_xfer.exs_evt_length));
		if (len < 0) {
			fprintf(stderr, "%s %s: %s\n", message,
					"asynch_wait send_queue",
					strerror(errno));
			n = 0;
			failed = 1;
			break;
		}

#if defined(ALLOW_PARANOID_CHECKING)
		if (completion.exs_evt_socket != fd) {
			fprintf(stderr, "%s got exs_evt_socket %d, "
				"expected %d\n",
				message, completion.exs_evt_socket, fd);
			n = 0;
			failed = 1;
			break;
		}
#endif

		if (sg_elements) {
			cb->total_bytes += completion.exs_evt_union
					.exs_evt_xferfile.exs_evt_length;
			xptr = (exs_xferfile_t *)completion.exs_evt_ahandle;
			if (get_debug_mask() & LOG_APPLICATION) {
				if (xptr->exs_xfv_type == EXS_IOVEC) {
					bufptr = xptr->exs_xfv.exs_iov.iov_base;
					buf_handle = xptr->
						exs_xfv.exs_iov.iov_mhandle;
				} else if (xptr->exs_xfv_type == EXS_FDVEC) {
					/* for now */
					bufptr = NULL;
					buf_handle = NULL;
				} else {/* for now */
					bufptr = NULL;
					buf_handle = NULL;
				}
				TRACE_DBG(LOG_APPLICATION,
					"%s starting iteration %u, fd %d, len %d, buffer address %p, buf_handle %p, send_queue %p, total_bytes %llu",
					message, i,
					fd, size, bufptr, buf_handle,
					cb->send_queue, cb->total_bytes);
			}
		} else {
			cb->total_bytes += completion.exs_evt_union
					.exs_evt_xfer.exs_evt_length;
			bufptr = completion.exs_evt_union
					.exs_evt_xfer.exs_evt_buffer;
			buf_handle = completion.exs_evt_union
					.exs_evt_xfer.exs_evt_mhandle;
			TRACE_DBG(LOG_APPLICATION,
				"%s starting iteration %u, fd %d, len %d, buffer address %p, buf_handle %p, send_queue %p",
				message, i, fd, size, bufptr,
				buf_handle, cb->send_queue);
		}
		} else {
			if (sg_elements) {
				xptr = &cb->xfertab[kk];
				
			} else {
				bufptr = cb->buftab[kk];
				buf_handle = cb->buf_mhantab[kk];
			}
			kk++;
			if (kk >= m)
				kk = 0;
		}	/* no queue */
		if (verify) {
			ptr = (char *)bufptr;
			endptr = &ptr[size];
			while (ptr != endptr) {
				if (nextch == 0x7f)
					nextch = ' ';
				*ptr++ = nextch++;
			}
		}
		if (sg_elements) {
			len = asynch_sendfile(fd, xptr, sg_elements,
							local_send_queue, xptr);
		} else {
			len = asynch_send(fd, (char *)bufptr, size,
					send_flags, buf_handle,
					local_send_queue);
		}
		if (len <= 0) {
			fprintf(stderr, "%s %s: %s\n", message,
					"asynch_send", strerror(errno));
			n = 0;
			failed = 1;
			break;
		}
		cb->total_transfers++;
	}	/* for */

	get_relative_time(&cb->start_time, &cb->secs, &cb->usecs, message);
	fprintf(stderr, "%s sent %u buffers, waiting for %u "
		"at %d.%06d\n", message, i, n, cb->secs, cb->usecs);

	/* when not waiting for implicit exs ACKs on the sends we just did,
	 * we have to send an additional message to tell server we are done
	 * before we can wait for him to tell us that he is also done
	 */
	 if (!failed && no_queue) {
		/* send "done" to server to say we're done */
		strcpy(text_buf, "no_queue done");
		len = strlen(text_buf);
		fprintf(stderr, "%s writeblock len %d \"%s\"\n",
			message, len, text_buf);
		len = writeblock(fd, text_buf, len, 0,
				EXS_MHANDLE_UNREGISTERED, cb->send_queue);
		if (len <= 0) {
			fprintf(stderr, "%s no_queue writeblock %s\n", message,
				strerror(errno));
		}
	}

	if(!failed && no_queue) {
		/* wait for server to tell us he is "done" */
		len = readpacket(fd, text_buf, TEXT_SIZE,
				EXS_MHANDLE_UNREGISTERED, cb->recv_queue);
		if (len <= 0) {
			fprintf(stderr, "%s readpacket %s\n", message,
				strerror(errno));
		} else {
			text_buf[len] = '\0';
			fprintf(stderr, "%s readpacket len %d \"%s\"\n",
				message, len, text_buf);
			len2 = strlen("done");
			if(len > len2 && (strncmp(text_buf,"done",len2) == 0)) {
				sscanf(text_buf+len2, "%llu %llu",
					&cb->total_transfers, &cb->total_bytes);
			}
		}
		/* send an "ok" to server */
		strcpy(text_buf, "ok");
		len = strlen(text_buf);
		fprintf(stderr, "%s writeblock len %d \"%s\"\n",
			message, len, text_buf);
		len = writeblock(fd, text_buf, len, 0,
				EXS_MHANDLE_UNREGISTERED, cb->send_queue);
		if (len <= 0) {
			fprintf(stderr, "%s writeblock %s\n", message,
				strerror(errno));
		}
	}


	if (!failed && no_queue == 0)
	{
	/* now wait for the last outstanding buffer(s) to complete */
	for ( ; n > 0; n--) {
		TRACE_DBG(LOG_APPLICATION,
			"%s calling asynch_wait, %u buffers left", message, n);
		len = asynch_wait(cb->send_queue, &completion, NULL);

		TRACE_DBG(LOG_APPLICATION,
			"%s got completion type %d, expected %d, errno %d, fd %d, length %zu",
				message, completion.exs_evt_type,
				sg_elements ? EXS_EVT_SENDFILE
						    : EXS_EVT_SEND,
				completion.exs_evt_errno,
				completion.exs_evt_socket,
				(sg_elements
					?  completion.exs_evt_union
					       .exs_evt_xferfile.exs_evt_length
					: completion.exs_evt_union
					       .exs_evt_xfer.exs_evt_length));
		if (len < 0) {
			fprintf(stderr, "%s %s: %s\n", message,
					"asynch_wait send_queue",
					strerror(errno));
			len = -1;
			break;
		}

#if defined(ALLOW_PARANOID_CHECKING)
		if (completion.exs_evt_socket != fd) {
			fprintf(stderr, "%s got exs_evt_socket %d, "
				"expected %d\n",
				message, completion.exs_evt_socket, fd);
			break;
		}
#endif

		if (sg_elements) {
			cb->total_bytes += completion.exs_evt_union
					.exs_evt_xferfile.exs_evt_length;
		} else {
			cb->total_bytes += completion.exs_evt_union
					.exs_evt_xfer.exs_evt_length;
		}
	}

	if (len < 0) {
		fprintf(stderr, "%s asynch %s\n", message,
			strerror(errno));
	}
	}	/* no queue */

	/* stop the timers immediately after last transaction completes */
	print_rates(cb, i, message);

	get_relative_time(&cb->start_time, &cb->secs, &cb->usecs, message);

	fprintf(stderr,"%s finished at %d.%06d\n",message,cb->secs,cb->usecs);
err1:
	/* start an alarm so that we will get aborted if we can't shut down */
	alarm(8);

	/* close the connection to the remote side */
	TRACE_DBG(LOG_APPLICATION,
		"%s exs_close(fd=%d), recv_queue %p",
			message, fd, cb->recv_queue);
	if(close_a_connection(cb, fd, message))
		goto err0;

	/* if we get all the way thru to here, we were completely successful */
	if (len >= 0) {
		level= 8;
	}
err0:
	free(text_buf);
	return dealloc_control_block(cb, level);
}	/* client_task */


/* here when we have just accepted a new client connection */
void *
server_thread(void *arg)
{
	int			len, client_fd;
	unsigned int		i, kk, m, n;
	struct sockaddr		*client;
	struct child_params	*param;
	struct sockaddr_storage	name;
	socklen_t		namelen;
	exs_event_t		completion;
	void			*bufptr;
	exs_mhandle_t		buf_handle;
	char			bufch, expectedch = '0';
	char			*text_buf;
	char 			*cptr;
	struct user_ahandle	*ahandle;
	char			message[100];
	exs_qhandle_t		local_recv_queue;
	struct control_block	*cb;

	if ((len = posix_memalign((void **)&text_buf, align, TEXT_SIZE + 2)) != 0) {
		fprintf(stderr, "Could not allocate initial buffer: %s\n",
				strerror(len));
	}

	param = (struct child_params *)arg;
	client_fd = param->client_fd;
	client = (struct sockaddr *)&param->client_addr;

	sprintf(message, "%s [%lu] %s fd %d:",
		my_name, pthread_self(), "agent", client_fd);

	fprintf(stderr, "%s Enter, fd %d\n", message, client_fd);

	TRACE_DBG(LOG_APPLICATION, "%s connected on fd %d to client %p",
			message, client_fd, client);

	/* we are now successfully connected to a remote client */
	if (client != NULL) {
		/* print address info about remote client */
		print_ip_address((struct sockaddr *)&param->client_addr,
					"remote client", message);
	}

	/* try out the added exs_getsockname() and exs_getpeername() functions*/
	namelen = sizeof(name);
	if (exs_getsockname(client_fd, (struct sockaddr *)&name, &namelen))
		fprintf(stderr, "%s %s: %s\n", message,
					"exs_getsockname", strerror(errno));
	print_ip_address((struct sockaddr *)&name, "server exs_getsockname",
			message);

	namelen = sizeof(name);
	if (exs_getpeername(client_fd, (struct sockaddr *)&name, &namelen))
		fprintf(stderr, "%s %s: %s\n", message,
					"exs_getpeername", strerror(errno));
	print_ip_address((struct sockaddr *)&name, "server exs_getpeername",
			message);

	exs_print_connection_attributes(stderr, client_fd, message);

	if (alloc_control_block(&cb, message) == EXIT_FAILURE) {
		free(text_buf);
		return NULL;
	}

	if (no_queue)
		local_recv_queue = NULL;
	else
		local_recv_queue = cb->recv_queue;

	/* wait for client to tell us to "go" */
	len = readpacket(client_fd, text_buf, TEXT_SIZE,
				EXS_MHANDLE_UNREGISTERED, cb->recv_queue);
	if (len <= 0) {
		fprintf(stderr,"%s readpacket %s\n", message,
					strerror(errno));
		goto error_out;
	}

	text_buf[2] = '\0';
	fprintf(stderr, "%s readpacket len %d \"%s\"\n",
		message, len, text_buf);
	if (len != (long)align) {
		fprintf(stderr, "%s got len %d, expected %d\n",
			message, len, 2);
		goto error_out;
	}

	/* repeatedly read a buffer from a client */
	fprintf(stderr, "%s reading %d-byte buffers %u times with "
		"%u active simultaneously\n", message, size, max_count, nbufs);

	/* m gets min(max_count, nbufs) */
	m = max_count;
	if (m > nbufs)
		m = nbufs;

	/* set up to recv the first nbufs buffers without waiting */
	for (i = 0, n = 0; i < m; i++, n++) {
		TRACE_DBG(LOG_APPLICATION,
			"%s starting iteration %u, fd %d, len %d, buffer address %p, buf_handle %p, local_recv_queue %p",
			message, i, client_fd, size, cb->buftab[n],
			cb->buf_mhantab[n], local_recv_queue);
		len = asynch_recv(client_fd, (char *)cb->buftab[n], size,
					recv_flags, local_recv_queue,
					cb->buf_ahantab[n],
					cb->buf_mhantab[n]);
		if (len <= 0)
			goto error_out;
		if (no_queue)
			cb->total_bytes += size;
	}

	/* start the timers immediately before first transaction starts */
	if( getrusage(RUSAGE_SELF, &cb->start_usage) < 0 )
		{
		fprintf(stderr, "%s getrusage start %s\n",
			message, strerror(errno));
		exit(EXIT_FAILURE);
		}
	if( clock_gettime(CLOCK_EXS, &cb->start_time) < 0 )
		{
		fprintf(stderr, "%s gettimeofday start %s\n",
			message, strerror(errno));
		exit(EXIT_FAILURE);
		}

	/* send an ack back to client to say we're ready */
	strcpy(text_buf, "ok");
	len = strlen(text_buf);
	fprintf(stderr, "%s writeblock len %d \"%s\"\n",
		message, len, text_buf);
	len = writeblock(client_fd, text_buf, align, 0,
				EXS_MHANDLE_UNREGISTERED, cb->send_queue);
	if (len <= 0) {
		fprintf(stderr,"%s writeblock len %d error %d: %s\n", message,
			len, errno, strerror(errno));
		goto error_out;
	}

	get_relative_time(&cb->start_time, &cb->secs, &cb->usecs, message);
	fprintf(stderr,"%s set up first %u buffers without waiting"
		" at %d.%06d\n", message, i, cb->secs, cb->usecs);

	len = 0;

	/* recv remaining packets after waiting for previous completion
	 * in order to reuse its buffer and buffer registration
	 */
	kk = 0;
	for ( ; i < max_count; i++) {
		if (no_queue == 0)
		{
		len = asynch_wait(cb->recv_queue, &completion, NULL);

		TRACE_DBG(LOG_APPLICATION,
			"%s got completion type %d, expected %d, errno %d, fd %d, length %zu",
			message, completion.exs_evt_type, EXS_EVT_RECV,
			completion.exs_evt_errno,
			completion.exs_evt_socket,
			completion.exs_evt_union.exs_evt_xfer.exs_evt_length);
		if (len < 0) {
			fprintf(stderr, "%s %s: %s\n", message,
					"asynch_wait recv_queue",
					strerror(errno));
			n = 0;
			break;
		}
		len = completion.exs_evt_union.exs_evt_xfer.exs_evt_length;
		if (socktype == SOCK_SEQPACKET && len < size) {
			fprintf(stderr, "%s %s: warning: i %d len %u < size %u\n",
					message, "asynch_wait recv_queue", i,
					len, size);
		}
		cb->total_bytes += completion.exs_evt_union
					.exs_evt_xfer.exs_evt_length;
		bufptr
			= completion.exs_evt_union.exs_evt_xfer.exs_evt_buffer;
		buf_handle
			= completion.exs_evt_union.exs_evt_xfer.exs_evt_mhandle;
		ahandle
			= completion.exs_evt_ahandle;

		TRACE_DBG(LOG_APPLICATION,
			"%s got message with length %zu: \"%c..%c\"",
				message,
				completion.exs_evt_union
					.exs_evt_xfer.exs_evt_length,
				((char *)bufptr)[0],
				((char *)bufptr)[completion.exs_evt_union
					.exs_evt_xfer.exs_evt_length - 1]);

		TRACE_DBG(LOG_APPLICATION,
			"%s bufptr %p, buf_handle %p, ahandle %p, recv_flags %0x",
			message, bufptr, buf_handle, ahandle, recv_flags);

		if (ahandle)
			TRACE_DBG(LOG_APPLICATION,
				"%s ahandle->length %d, ahandle->mhandle %p, ahandle->address %p, bufptr %p",
				message, (int)ahandle->length, ahandle->mhandle,
				ahandle->address, bufptr);

		if (verify) {
			cptr = (char *)bufptr;
			TRACE_DBG(LOG_APPLICATION,
				"buffer: \"%c...%c\"\n",
				cptr[0], cptr[len - 1]);
			bufch = *cptr;
			if (bufch && (expectedch != bufch)) {
				fprintf(stderr,
					"len=%d i=%llu j=%td expected '%c' got '%c'\n",
					len, cb->total_transfers, (cptr - (char *)bufptr),
					expectedch, bufch);
				TRACE_DBG(LOG_APPLICATION,
					"full buffer: \"%s\"\n",
					cptr);
				break;
			}
			expectedch = cptr[len - 1] + 1;
			if (!isprint(expectedch)) {
				expectedch = ' ';
			}
		}

		} else {
			cb->total_bytes += size;
			bufptr = cb->buftab[kk];
			buf_handle = cb->buf_mhantab[kk];
			ahandle = cb->buf_ahantab[n];
			kk++;
			if (kk >= m)
				kk = 0;
		}	/* no queue */
		TRACE_DBG(LOG_APPLICATION,
			"%s starting iteration %u, fd %d, len %d, buffer address %p, buf_handle %p, local_recv_queue %p",
			message, i, client_fd, size, bufptr,
			buf_handle, local_recv_queue);
		//usleep(10000);
		len = asynch_recv(client_fd, (char *)bufptr, size,
				recv_flags,
				local_recv_queue, ahandle, buf_handle);
		if (len != size) {
			n = 0;
			break;
		}
		cb->total_transfers++;
	}	/* for */

	get_relative_time(&cb->start_time, &cb->secs, &cb->usecs, message);
	fprintf(stderr, "%s set up for %u buffers, waiting for"
		" %u at %d.%06d\n", message, i, n, cb->secs, cb->usecs);

	/* now wait for the last outstanding buffers to complete */
	if (no_queue)
	{	/* when we are not getting any exs ACK notifies,
		 * we depend on client to send us an explicit message
		 */
		len = readpacket(client_fd, text_buf, TEXT_SIZE,
				EXS_MHANDLE_UNREGISTERED, cb->recv_queue);
		if (len <= 0) {
			fprintf(stderr,
				"%s no_queue readpacket %s\n",
				message, strerror(errno));
		} else {
			text_buf[len] = '\0';
			fprintf(stderr, "%s readpacket len %d "
				"\"%s\"\n", message, len, text_buf);
		}
	}
	else
	{
	for ( ; n > 0; n--) {
		len = asynch_wait(cb->recv_queue, &completion, NULL);

		TRACE_DBG(LOG_APPLICATION,
			"%s got completion type %d, expected %d, errno %d, fd %d, length %zu",
			message, completion.exs_evt_type, EXS_EVT_RECV,
			completion.exs_evt_errno,
			completion.exs_evt_socket,
			completion.exs_evt_union.exs_evt_xfer.exs_evt_length);
		if (len < 0) {
			fprintf(stderr, "%s %s: %s\n", message,
					"asynch_wait recv_queue",
					strerror(errno));
			break;
		}
		cb->total_bytes += completion.exs_evt_union
						.exs_evt_xfer.exs_evt_length;
		bufptr
			= completion.exs_evt_union.exs_evt_xfer.exs_evt_buffer;
		buf_handle
			= completion.exs_evt_union.exs_evt_xfer.exs_evt_mhandle;
		ahandle
			= completion.exs_evt_ahandle;

		if (verify) {
			cptr = (char *)bufptr;
			len = completion.exs_evt_union.exs_evt_xfer.exs_evt_length;
			TRACE_DBG(LOG_APPLICATION,
				"buffer: \"%c...%c\"\n",
				cptr[0], cptr[len - 1]);
			bufch = *cptr;
			if (bufch && (expectedch != bufch)) {
				fprintf(stderr,
					"len=%d i=%d j=%td expected '%c' got '%c'\n",
					len, i, (cptr - (char *)bufptr),
					expectedch, bufch);
				TRACE_DBG(LOG_APPLICATION,
					"full buffer: \"%s\"\n",
					cptr);
				break;
			}
			expectedch = cptr[len - 1] + 1;
			if (!isprint(expectedch)) {
				expectedch = ' ';
			}
		}

		cb->total_transfers++;
	}	/* for */
	}

	if (len < 0) {
		fprintf(stderr, "%s asynch %s\n", message,
			strerror(errno));
	}
	TRACE_DBG(LOG_APPLICATION, "%s finished, total_bytes %llu",
			message, cb->total_bytes);

	/* when client's send was "persistent", we need to send him a message
	 * so he can send us back a message to break the persistent send
	 * @@@@@
	 */
	if(no_queue) {
		/* send a message back to client to say we're done */
		snprintf(text_buf, TEXT_SIZE, "done %llu %llu",
			cb->total_transfers, cb->total_bytes);
		len = strlen(text_buf);
		fprintf(stderr, "%s writeblock len %d \"%s\"\n",
				message, len, text_buf);
		len = writeblock(client_fd, text_buf, len, 0,
				EXS_MHANDLE_UNREGISTERED, cb->send_queue);
		if (len <= 0) {
			fprintf(stderr,"%s writeblock %s\n",
				message, strerror(errno));
		}

		bufptr = cb->buftab[0];
		buf_handle = cb->buf_mhantab[0];
		ahandle = cb->buf_ahantab[0];
		for (kk = 1; ; kk++) {
			len = asynch_recv(client_fd, (char *)bufptr, size,
					recv_flags, cb->recv_queue, ahandle,
					buf_handle);
			if (len != size) {
				fprintf(stderr, "%s asynch_recv len %d, "
					"expected %d\n", message, len, size);
				break;
			}
			len = asynch_wait(cb->recv_queue, &completion, NULL);
			if (len < 0){
				fprintf(stderr, "%s %s: %s\n", message,
						"asynch_wait recv_queue "
						"after done", strerror(errno));
				break;
			}
			len = completion.exs_evt_union
						.exs_evt_xfer.exs_evt_length;
			/* see if client broke his persistent send yet */
			if (strncmp((char *)bufptr, "ok", 2) == 0) {
				fprintf(stderr,"%s recv %u len %d got \"ok\"\n",
					message, kk, len);
				break;
			}
			TRACE_DBG(LOG_APPLICATION, "%s recv %u len %d from client to break persistent send,",
					message, kk, len);
		}
	}

	/* stop the timers immediately after last transaction completes */
	print_rates(cb, i, message);

#if 0
	/* now try to receive 1 last message (which should never arrive) */
	fprintf(stderr, "%s starting final recv that should never arrive\n",
		message);
	len = asynch_recv(client_fd, (char *)buftab[0], size,
					recv_flags, recv_queue,
					buf_ahantab[0],
					buf_mhantab[0]);

	/* if recv was posted ok, wait for 2 seconds for a FLUSH_ERR */
	if (len != size) {
		fprintf(stderr, "%s unable to start final recv\n", message);
	} else {
		struct timeval	timeout;

		timeout.tv_sec = 2;
		timeout.tv_usec = 0;
		fprintf(stderr, "%s starting wait for final recv\n", message);
		len = asynch_wait(recv_queue, &completion, &timeout);
		if (len < 0) {
			fprintf(stderr, "%s %s: %s\n", message,
					"asynch wait recv_queue",
					strerror(errno));
		} else {
			fprintf(stderr, "%s final wait returned len %d\n",
				message, len);
		}
	}
#endif

	/* delay to give client time to call exs_close() first */
//	fprintf(stderr,"%s giving client 1.5 seconds to call exs_close()\n",
//								message);
//	small_delay.tv_sec = 1;
//	small_delay.tv_nsec = 500000000;
//	if (nanosleep(&small_delay, NULL))
//			fprintf(stderr, "%s %s: %s\n", message,
//				"nanosleep small delay", strerror(errno));

error_out:
	/* close the server connection to the remote client */
	TRACE_DBG(LOG_APPLICATION,
		"%s exs_close(client_fd=%d), recv_queue %p",
		message, client_fd, cb->recv_queue);

	/*****
	exs_set_debug_flag(1);
	*****/

	/* start an alarm so that we will get aborted if we can't shut down */
	alarm(8);

	close_a_connection(cb, client_fd, message);

	alarm(0);

	get_relative_time(&cb->start_time, &cb->secs, &cb->usecs, message);
	fprintf(stderr,"%s finished at %d.%06d\n",message,cb->secs,cb->usecs);

	fprintf(stderr, "%s Leave, fd %d\n", message, client_fd);
	free(arg);
	dealloc_control_block(cb, 8);

	exs_print_counters(stderr);
	fflush(stdout);

	/* Before we exit, let the server know that we're exiting */
	pthread_mutex_lock(&exit_mutex);
	if (--nthreads == 0) {
		pthread_cond_signal(&exit_cond);
	}
	pthread_mutex_unlock(&exit_mutex);

	free(text_buf);
	return NULL;
}	/* server_thread */


/* returns 0 on success, -1 on error */
int
listen_task(void)
{
	int			listening_fd, nevents, ret = -1, len;
	exs_qhandle_t		accept_queue;
	struct sockaddr_storage	server_addr;
	exs_acceptaddr_t	address_vector[ADDRESS_ITEMS];
	exs_event_t		accept_vector[LISTENING_DEPTH];
	struct child_params	*param;
	char			message[100];
	pthread_t		kid_id;
	exs_event_t		completion;

	sprintf(message, "%s [%lu] %s:",
		my_name, pthread_self(), __func__);

	print_sock_sizes(message);

	fprintf(stderr, "%s Enter, nloops %u\n", message, n_listener_loops);

	/* need to create an event_queue for accepting connects to server */
	accept_queue = exs_qcreate(LISTENING_DEPTH);
	if (accept_queue == EXS_QHANDLE_INVALID) {
		perror("server exs_qcreate");
		ret = -1;
		goto err0;
	}

	/* establish a server "listening post" */
	listening_fd = openserver(server_port, server_name, LISTENING_DEPTH,
				socktype, (struct sockaddr *)&server_addr,
				&our_conn_params);
	TRACE_DBG(LOG_APPLICATION,
		"%s openserver returned listening_fd %d, accept_queue %p, depth %d",
		message, listening_fd, accept_queue, LISTENING_DEPTH);
	if (listening_fd < 0) {
		fprintf(stderr, "%s %s: %s\n", message,
				"open listening post", strerror(errno));
		goto err1;
	}

	/* we are now successfully established as a server */
	/* infinite loop -- server serves forever, must be killed by ^C */
	for ( ; n_listener_loops-- > 0; ) {

		print_ip_address((struct sockaddr *)&server_addr,
				"listening", message);
		print_attributes(listening_fd, message);

		/* set up the address vector so exs_accept will return the
		 * peer address of the new client
		 */
		param = calloc(1, sizeof(*param));
		if (param == NULL) {
			perror("listener calloc struct child_params");
			break;
		}
		address_vector[0].exs_addr=(struct sockaddr*)&param->client_addr;
		address_vector[0].exs_addrlen = sizeof(param->client_addr);
		address_vector[0].exs_ahandle = NULL;

		TRACE_DBG(LOG_APPLICATION,
			"%s exs_addr %p, exs_addrlen %d, exs_ahandle %p",
			message, address_vector[0].exs_addr,
			address_vector[0].exs_addrlen,
			address_vector[0].exs_ahandle);

		/* accept a client connection (block until one arrives) */
		if (exs_accept(listening_fd, address_vector, ADDRESS_ITEMS, 0,
				accept_queue) < 0 ) {
			fprintf(stderr, "%s %s: %s\n", message,
					"exs_accept", strerror(errno));
			break;
		}
		TRACE_DBG(LOG_APPLICATION, "%s exs_accept returned ok",
					message);

		/* in exs need to wait for accept to complete asynchronously */
		nevents = exs_qdequeue(accept_queue, accept_vector,
					ADDRESS_ITEMS, NULL);
		TRACE_DBG(LOG_APPLICATION,
			"%s exs_qdequeue nevents=%d accept_vector[0] type %d is %s",
			message, nevents, accept_vector[0].exs_evt_type,
			accept_vector[0].exs_evt_type == EXS_EVT_ACCEPT
					? "ACCEPT" : "not ACCEPT");
		if (nevents < 0) {
			fprintf(stderr, "%s %s: %s\n", message,
					"exs_qdequeue", strerror(errno));
			break;
		}
		if (nevents != 1) {
			fprintf(stderr, "%s exs_qdequeue dequeued "
					"%d items, expected 1\n", message,
					nevents);
			break;
		}
		if (accept_vector[0].exs_evt_type != EXS_EVT_ACCEPT) {
			fprintf(stderr, "%s exs_qdequeue dequeued "
					"item type %d, expected %d\n", message,
					accept_vector[0].exs_evt_type,
					EXS_EVT_ACCEPT);
			break;
		}
		if (accept_vector[0].exs_evt_errno != 0) {
			errno = accept_vector[0].exs_evt_errno;
			fprintf(stderr, "%s %s: %s\n", message,
				"exs_qdequeue status", strerror(errno));
			break;
		}
		param->client_fd = accept_vector[0].exs_evt_union.exs_evt_accept
				.exs_evt_new_socket;
		TRACE_DBG(LOG_APPLICATION,
			"%s exs_evt_new_socket %d, new_client %p",
			message, param->client_fd, &param->client_addr);

		/* dump out table of first 10 fd's */
		if (get_debug_mask() & LOG_APPLICATION)
			fstatdump(10);

		pthread_mutex_lock(&exit_mutex);
		++nthreads;
		pthread_mutex_unlock(&exit_mutex);

		errno = pthread_create(&kid_id, NULL, server_thread, param);
		if (errno != 0) {
			fprintf(stderr, "%s %s: %s\n", message,
				"pthread_create", strerror(errno));

			pthread_mutex_lock(&exit_mutex);
			--nthreads;
			pthread_mutex_unlock(&exit_mutex);
		} else {
			pthread_detach(kid_id);
			TRACE_DBG(LOG_APPLICATION,
				"*** %s child with thread_id %lu started",
				message, (long)kid_id);
		}
	}	/* for ever */

	pthread_mutex_lock(&exit_mutex);
	while (nthreads != 0) {
		pthread_cond_wait(&exit_cond, &exit_mutex);
	}
	pthread_mutex_unlock(&exit_mutex);

	fflush(stdout);

	/* if we get all the way thru to here, we were completely successful */
	ret = EXIT_SUCCESS;

err1:
	if ((len = exs_close(listening_fd, 0, accept_queue, NULL)) < 0) {
		fprintf(stderr, "%s fd %d: %s\n", message,
			listening_fd, strerror(errno));
	} else {
		TRACE_DBG(LOG_APPLICATION,
			"%s fd %d: call asynch_wait", message, listening_fd);
		len = asynch_wait(accept_queue, &completion, NULL);
		if (len < 0) {
			fprintf(stderr, "%s fd %d: %s\n", message,
					listening_fd, strerror(errno));
			len = -1;
		}
	}
	TRACE_DBG(LOG_APPLICATION,
		"%s exs_qdelete accept_queue %p", message, accept_queue);
	if (exs_qdelete(accept_queue) < 0) {
		fprintf(stderr, "%s %s: %s\n", message,
			"exs_qdelete accept_queue", strerror(errno));
	}

	fprintf(stderr, "%s Leave, return %d\n",
		message, ret);
err0:
	return ret;
}	/* listen_task */


int
main(int argc, char *argv[])
{
	int	c, n, ret = EXIT_FAILURE;
	unsigned int		debug_mask;
	unsigned int		len;
	unsigned int		u;
	/* sizeof(struct sockaddr_in)	is 16
	 * sizeof(struct sockaddr_in6)	is 28
	 * sizeof(struct sockaddr)	is 16 !!! should be 28 !!!
	 */
	char			*subopt_ptr, *subopt_value;
	char			*ptr;
	time_t			now;
	char			message[100];
//	struct timespec		small_delay;

	server_name = NULL;
	server_port = DEFAULT_PORT;
	max_count = DEFAULT_COUNT;
	size = DEFAULT_SIZE;
	nbufs = DEFAULT_NBUFS;
	n_listener_loops = DEFAULT_N_LISTENER_LOOPS;
	debug_mask = 0U;
	verify = 0;
	our_conn_params.fd_flags = 0;
	our_conn_params.small_packet_max_size = INT_MAX;
	our_conn_params.max_inline = INT_MAX;
	our_conn_params.comp_cpu = INT_MAX;
	our_conn_params.max_credits = INT_MAX;
	our_conn_params.stream_buf_size = INT_MAX;

	sprintf(message, "%s [%lu] %s:",
		my_name, pthread_self(), __func__);

	time(&now);
	fprintf(stderr, "%s Enter on %s\n", message, ctime(&now));

	while ((c = getopt(argc, argv, OPTIONS)) != EOF) {
		switch (c) {
		case 't':
		case 'T':
			/* use SOCK_STREAM (TCP) semantics
			 * (default is SOCK_SEQPACKET)
			 */
			socktype = SOCK_STREAM;
			break;
		case 'y':
		case 'Y':
			verify = 1;
			break;
		case 'n':
		case 'N':
			/* turn off pseudo-nagle algorithm */
			our_conn_params.fd_flags |= EXS_FD_NODELAY;
			break;
		case 'q':
		case 'Q':
			/* don't wait in notify queue after send/recv */
			no_queue = 1;
			break;
		case 'x':
		case 'X':
			/* use "expert mode" to bypass parameter checking */
			u = exs_get_global_flags();
			exs_set_global_flags(u | EXS_GLOBAL_BYPASS_PARAM_CHECKING);
			break;
		case 'k':
		case 'K':
			/* copy unregistered immediate data into our buffer */
			our_conn_params.fd_flags |= EXS_FD_SP_COPY_UNREG;
			break;
		case 'm':
		case 'M':
			/* set max_size of "small packet" immediate data */
			errno = 0;
			n = strtol(optarg, &ptr, 0);
			if (ptr == optarg || errno != 0 || n < 0
			|| n >= (64 * 1024 * 1024)
			|| strspn(ptr, WHITE_SPACE) != strlen(ptr)) {
				fprintf(stderr, "%s illegal -%c value "
					"\"%s\"\n", message, c, optarg);
				exit(EXIT_FAILURE);
			}
			our_conn_params.small_packet_max_size = n;
			break;
		case 'g':
		case 'G':
			/* number of scatter_gather elements in sendfile */
			errno = 0;
			n = strtol(optarg, &ptr, 0);
			if (ptr == optarg || errno != 0 || n <= 0
			|| strspn(ptr, WHITE_SPACE) != strlen(ptr)) {
				fprintf(stderr, "%s illegal -%c value "
					"\"%s\"\n", message, c, optarg);
				exit(EXIT_FAILURE);
			}
			sg_elements = n;
			break;
		case 'u':
		case 'U':
			unregistered_buffers_flag = 1;
			break;
		case 'b':
		case 'B':
			/* number simultaneously active buffers */
			errno = 0;
			u = strtoul(optarg, &ptr, 0);
			if (ptr == optarg || errno != 0 || u == 0
			|| strspn(ptr, WHITE_SPACE) != strlen(ptr)) {
				fprintf(stderr, "%s illegal -%c value "
					"\"%s\"\n", message, c, optarg);
				exit(EXIT_FAILURE);
			}
			nbufs = u;
			break;
		case 'l':
		case 'L':
			/* number listener loops */
			errno = 0;
			u = strtoul(optarg, &ptr, 0);
			if (ptr == optarg || errno != 0 || u == 0
			|| strspn(ptr, WHITE_SPACE) != strlen(ptr)) {
				fprintf(stderr, "%s illegal -%c value "
					"\"%s\"\n", message, c, optarg);
				exit(EXIT_FAILURE);
			}
			n_listener_loops = u;
			break;
		case 'i':
			input_flag = 1;
			break;
		case 'o':
		case 'O':
			output_flag = 1;
			break;
		case 'D':
			u = exs_get_global_flags();
			exs_set_global_flags(u | EXS_GLOBAL_TRACE);
			debug_mask |= LOG_UTILITIES;
		case 'd':
			debug_mask |= LOG_APPLICATION;
			break;
		case 'I':
			errno = 0;
			u = strtoul(optarg, &ptr, 0);
			if (ptr == optarg || errno != 0
			|| strspn(ptr, WHITE_SPACE) != strlen(ptr)) {
				fprintf(stderr, "%s illegal -%c value "
					"\"%s\"\n", message, c, optarg);
				exit(EXIT_FAILURE);
			}
			our_conn_params.max_inline = u;
			break;
		case 'a':
		case 'A':
			/* IP name or address of server */
			len = strlen(optarg);
			if (len <= 0 || strspn(optarg, WHITE_SPACE) == len)
				break;
			server_name = malloc(len+1);
			if (server_name == NULL) {
				fprintf(stderr, "%s %s: %s\n", message,
					"malloc for -a", strerror(errno));
				break;
			}
			strcpy(server_name, optarg);
			break;
		case 'p':
		case 'P':
			/* port number */
			len = strlen(optarg);
			if (len <= 0 || strspn(optarg, WHITE_SPACE) == len)
				break;
			server_port = malloc(len+1);
			if (server_port == NULL) {
				fprintf(stderr, "%s %s: %s\n", message,
					"malloc for -p", strerror(errno));
				break;
			}
			strcpy(server_port, optarg);
			break;
		case 'c':
		case 'C':
			/* count of number of iterations */
			errno = 0;
			u = strtoul(optarg, &ptr, 0);
			if (ptr == optarg || errno != 0 || u == 0
			|| strspn(ptr, WHITE_SPACE) != strlen(ptr)) {
				fprintf(stderr, "%s illegal -%c value "
					"\"%s\"\n", message, c, optarg);
				exit(EXIT_FAILURE);
			}
			max_count = u;
			break;
		case 's':
		case 'S':
			/* size (in bytes) of data block sent in message */
			errno = 0;
			n = strtol(optarg, &ptr, 0);
			if (ptr == optarg || errno != 0 || n <= 0
			|| strspn(ptr, WHITE_SPACE) != strlen(ptr)) {
				fprintf(stderr, "%s illegal -%c value "
					"\"%s\"\n", message, c, optarg);
				exit(EXIT_FAILURE);
			}
			size = n;
			break;
		case 'e':
		case 'E':
			/* force busy-polling instead of event notification */
			our_conn_params.fd_flags |= EXS_FD_BUSYPOLL;
			break;
		case 'f':
		case 'F':
			errno = 0;
			n = strtol(optarg, &ptr, 0);
			if (ptr == optarg || errno != 0 || n <= 0
			|| strspn(ptr, WHITE_SPACE) != strlen(ptr)) {
				fprintf(stderr, "%s illegal -%c value "
					"\"%s\"\n", message, c, optarg);
				exit(EXIT_FAILURE);
			}
			our_conn_params.comp_cpu = n;
			break;
		case 'r':
		case 'R':
			errno = 0;
			n = strtol(optarg, &ptr, 0);
			if (ptr == optarg || errno != 0 || n <= 0
			|| strspn(ptr, WHITE_SPACE) != strlen(ptr)) {
				fprintf(stderr, "%s illegal -%c value "
					"\"%s\"\n", message, c, optarg);
				exit(EXIT_FAILURE);
			}
			our_conn_params.max_credits = n;
			break;
		case 'z':
		case 'Z':
			errno = 0;
			n = strtol(optarg, &ptr, 0);
			if (ptr == optarg || errno != 0 || n < 0
			|| strspn(ptr, WHITE_SPACE) != strlen(ptr)) {
				fprintf(stderr, "%s illegal -%c value "
					"\"%s\"\n", message, c, optarg);
				exit(EXIT_FAILURE);
			}
			if (n == 0 && (our_conn_params.fd_flags
						& EXS_FD_NO_STREAM_ADVERT)) {
				fprintf(stderr, "Must have stream buffer size > 0 with no_stream_advert\n");
				exit(EXIT_FAILURE);
			}
			our_conn_params.stream_buf_size = n;
			break;
		case 'w':
		case 'W':
			subopt_ptr = optarg;
			while (*subopt_ptr != '\0') {
				switch (getsubopt(&subopt_ptr, suboptions, &subopt_value)) {
				case SUBOPT_NO_STREAM_ADVERT:
					if (our_conn_params.stream_buf_size == 0) {
						fprintf(stderr, "Must have stream buffer size > 0 with no_stream_advert\n");
						exit(EXIT_FAILURE);
					}
					our_conn_params.fd_flags |= EXS_FD_NO_STREAM_ADVERT;
					break;
				default:
					fprintf(stderr, "Unknown suboption in ``%s''\n", optarg);
					exit(EXIT_FAILURE);
				}
			}
			break;
		case 'h':
		case 'H':
			fprintf(stderr, "%s version %d help info\n",
				message, EXS_VERSION);
			fprintf(stderr, "command line: ./pumpexs options\n");
			fprintf(stderr, "-a <server's ip-address or dns-name>\n");
			fprintf(stderr, "-b <number>\tnumber of simultaneously outstanding transmissions, default 1\n");
			fprintf(stderr, "-c <count>\tnumber of transmissions to perform\n");
			fprintf(stderr, "-d\tturn on pumpexs debug trace output\n");
			fprintf(stderr, "-D\tturn on pumpexs and exs debug trace output\n");
			fprintf(stderr, "-e\tuse busy-polling instead of event notification\n");
			fprintf(stderr, "-f <cpu>\tpin the exs completion thread to this cpu\n");
			fprintf(stderr, "-g <number>\tnumber of sendfile scatter_gather elements\n");
			fprintf(stderr, "-i\trun as the input receiver (server)\n");
			fprintf(stderr, "-I <number>\tmaximum amount of inline data\n");
			fprintf(stderr, "-k\tcopy unregistered data in exs\n");
			fprintf(stderr, "-l <number>\tnumber of listener accepts\n");
			fprintf(stderr, "-m <maximum>\tmaximum size in bytes of data to copy, default 0\n");
			fprintf(stderr, "-n\tdon't use pseudo-nagle algorithm in exs\n");
			fprintf(stderr, "-o\trun as the output sender (client)\n");
			fprintf(stderr, "-p <server's port number>\n");
			fprintf(stderr, "-q\tdon't use notification queues\n");
			fprintf(stderr, "-r\tmaximum flow control credits\n");
			fprintf(stderr, "-s <size>\tnumber of bytes in each transmission\n");
			fprintf(stderr, "-t\tuse SOCK_STREAM (TCP) semantics (default SOCK_SEQPACKET)\n");
			fprintf(stderr, "-u\tuse unregistered buffers\n");
			fprintf(stderr, "-w\textra suboptions, see below\n");
			fprintf(stderr, "-x\tbypass parameter checking in exs\n");
			fprintf(stderr, "-y\tsend repeating pattern and verify all received data\n");
			fprintf(stderr, "-z\tmaximum size of SOCK_STREAM receive buffer\n");
			fprintf(stderr, "\nSuboptions:\n");
			fprintf(stderr, "\tno_stream_advert\tdisallow recv buffer advertisements\n");
			exit(EXIT_SUCCESS);
			break;
		default:
			fprintf(stderr, "%s unknown option '%c'\n", message, c);
			exit(EXIT_FAILURE);
		}	/* switch */
	}	/* while */

	if (input_flag == output_flag) {
		if (input_flag)
			fprintf(stderr, "%s must specify either -i or -o,"
				" not both\n", message);
		else
			fprintf(stderr, "%s must specify either -i or -o"
				"\n", message);
		exit(EXIT_FAILURE);
	}

	if (size <= 0) {
		fprintf(stderr, "%s %d-bytes too big for integer\n",
				message, size);
		exit(EXIT_FAILURE);
	}

	/* for now */
	if (input_flag && sg_elements > 0) {
		fprintf(stderr, "%s cannot use -g together with -i\n", message);
		exit(EXIT_FAILURE);
	}

	if( (our_conn_params.fd_flags & EXS_FD_SP_COPY_UNREG)
					&& unregistered_buffers_flag == 0) {
		fprintf(stderr, "%s ignoring -k switch when no -u switch\n",
					message);
		our_conn_params.fd_flags &= ~EXS_FD_SP_COPY_UNREG;
	}

	if (output_flag && n_listener_loops != DEFAULT_N_LISTENER_LOOPS) {
		fprintf(stderr, "%s ignoring -l switch with -o switch\n",
					message);
	}

	if (socktype == SOCK_SEQPACKET) {
		/* exs silently turns off nagle algorithm when SOCK_SEQPACKET */
		our_conn_params.fd_flags |= EXS_FD_NODELAY;

		/* stream buffer not supported with SOCK_SEQPACKET */
		our_conn_params.stream_buf_size = INT_MAX;
	}

	set_debug_mask(debug_mask);

	/* ignore sigpipe signals -- handle the error synchronously */
	no_sigpipe();

	/* initialize the exs API */
	if (exs_init(EXS_VERSION) < 0) {
		fprintf(stderr, "%s %s: %s\n", message,
					"exs_init", strerror(errno));
		exit(EXIT_FAILURE);
	}

	fprintf(stderr, "%s -%c %36s \"%s\"\n",
		message, 'a', "server_name", server_name);
	fprintf(stderr, "%s -%c %36s \"%s\"\n",
		message, 'p', "server_port", server_port);
	fprintf(stderr, "%s -%c %36s %u\n",
		message, 'c', "count", max_count);
	fprintf(stderr, "%s -%c %36s %d\n",
		message, 's', "size", size);
	fprintf(stderr, "%s -%c %36s %d\n",
		message, 'm', "max small packet size",
		our_conn_params.small_packet_max_size);
	fprintf(stderr, "%s -%c %36s %d\n",
		message, 'I', "max inline", our_conn_params.max_inline);
	fprintf(stderr, "%s -%c %36s %d\n",
		message, 'F', "completion thread cpu", our_conn_params.comp_cpu);
	fprintf(stderr, "%s -%c %36s %u\n",
		message, 'b', "nbufs", nbufs);
	fprintf(stderr, "%s -%c %36s %d\n",
		message, 'g', "sg_elements", sg_elements);
	fprintf(stderr, "%s -%c %36s \"%s\"\n",
		message, 't', "socket type",
		socktype == SOCK_SEQPACKET ? "SOCK_SEQPACKET" : "SOCK_STREAM");
	if (socktype == SOCK_STREAM) {
		fprintf(stderr, "%s -%c %36s \"%s\"\n",
			message, 'n', "use nagle algorithm",
			our_conn_params.fd_flags & EXS_FD_NODELAY ? "no" : "yes");
	}
	fprintf(stderr, "%s -%c %36s \"%s\"\n",
		message, 'q', "wait in notify queue after send/recv",
						no_queue == 0 ? "yes" : "no");
	fprintf(stderr, "%s -%c %36s \"%s\"\n",
		message, 'u', "use unregistered buffers",
			unregistered_buffers_flag ? "yes" : "no");
	if (unregistered_buffers_flag) {
		fprintf(stderr, "%s -%c %36s \"%s\"\n",
			message, 'k', "copy unregistered data on send",
			our_conn_params.fd_flags & EXS_FD_SP_COPY_UNREG_ON_SEND
								? "yes" : "no");
		fprintf(stderr, "%s -%c %36s \"%s\"\n",
			message, 'k', "copy unregistered data on recv",
			our_conn_params.fd_flags & EXS_FD_SP_COPY_UNREG_ON_RECV
								? "yes" : "no");
	}
	fprintf(stderr, "%s -%c %36s \"%s\"\n",
		message, 'x', "bypass exs parameter checking",
			our_conn_params.fd_flags & EXS_FD_BYPASS_PARAM_CHECKING
								? "yes" : "no");
	if (input_flag ) {
		fprintf(stderr, "%s -%c %36s %d\n",
			message, 'l', "n_listener_loops", n_listener_loops);
	}

	if (output_flag) {
		ret = client_task();
		/* delay to give server time to call exs_close() first */
//		fprintf(stderr,"%s giving server 1.5 seconds to call "
//			"exs_close()\n", message);
//		small_delay.tv_sec = 1;
//		small_delay.tv_nsec = 500000000;
//		if (nanosleep(&small_delay, NULL))
//			fprintf(stderr, "%s %s: %s\n", message,
//				"nanosleep small delay", strerror(errno));

	} else {
		ret = listen_task();
	}

	exs_print_counters(stderr);
	exs_print_calloc_report(stderr);

	fprintf(stderr, "%s Leave with %s\n", message,
		ret == EXIT_SUCCESS ? "SUCCESS" : "FAILURE");
	return ret;
}	/* main */

/* vi: set autoindent tabstop=8 shiftwidth=8 : */
